/*
 * Copyright (C) 2014, Stephan Mueller <smueller@chronox.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU General Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, ALL OF
 * WHICH ARE HEREBY DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF NOT ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef CRYPTOPERF_H
#define CRYPTOPERF_H

#include <linux/slab.h>
#include <linux/scatterlist.h>
#include <linux/random.h>
#include <linux/crypto.h>
#include <crypto/hash.h>

#define DRIVER_NAME "cryptoperf"

#if 1
#define dbg(fmt, ...) pr_info(DRIVER_NAME": " fmt, ##__VA_ARGS__)
#else
#define dbg(fmt, ...)
#endif

/* default execution time for a test in seconds */
#define DFLT_EXECTIME 10

/*
 * Cipher-specific test data
 */
struct rng_def {
	unsigned int blocksize; /* config option */
	unsigned char *scratchpad;
	struct crypto_rng *rng;
};

struct sdesc {
	struct shash_desc shash;
	char ctx[];
};

struct shash_def {
	unsigned int hmac;	/* config option */
	unsigned int digestsize;/* config option */
	unsigned char *scratchpad;
	struct sdesc *desc;
};

struct blkcipher_def {
	unsigned int keysize;	/* config option */
	unsigned int blocksize; /* config option */
	unsigned char *scratchpad;
	struct scatterlist sg;
	struct blkcipher_desc desc;
};

struct tcrypt_result {
	struct completion completion;
	int err;
};

struct ablkcipher_def {
	unsigned int keysize;	/* config option */
	unsigned int blocksize; /* config option */
	unsigned char *scratchpad;
	unsigned char *ivdata;
	struct scatterlist sg;
	struct crypto_ablkcipher *tfm;
	struct ablkcipher_request *req;
	struct tcrypt_result result;
};

/*
 * Test result data
 */
struct cp_res {
	__u64 rounds;
	__u64 processed_bytes;
	__u64 totaltime;
	unsigned int byteperop;
};

/*
 * Test case definition
 */
struct cp_test;
struct cp_test {
	char *testname;
	char *driver_name;
	char *type;
	unsigned int exectime;
	struct cp_res results;
	int (*init_test)(struct cp_test *test);
	unsigned int (*exec_test)(struct cp_test *test);
	void (*fini_test)(struct cp_test *test);

	/* information not to be set by user */
	union {
		struct rng_def rng;
		struct blkcipher_def blkcipher;
		struct ablkcipher_def ablkcipher;
		struct shash_def shash;
	} u;
	struct list_head list;
};

static inline __u64 cp_ts2u64(struct timespec *ts)
{
	return ((ts->tv_sec << 32) | ts->tv_nsec);
}

/*
 * This is x86 specific to reduce the CPU jitter
 */
static void __always_inline cp_cpusetup(void)
{
#ifdef CONFIG_X86
	asm volatile("cpuid"::"a" (0), "c" (0): "memory");
	asm volatile("cpuid"::"a" (0), "c" (0): "memory");
	asm volatile("cpuid"::"a" (0), "c" (0): "memory");
#endif
}

static __always_inline void cp_get_nstime(struct timespec *ts)
{
	*ts = current_kernel_time();
}

static __always_inline void cp_start_time(struct timespec *ts)
{
	cp_cpusetup();
	cp_get_nstime(ts);
}

static __always_inline void cp_end_time(struct timespec *ts)
{
	cp_get_nstime(ts);
}

static inline void *cp_zalloc(size_t len)
{
	return kzalloc(len, GFP_KERNEL);
}
static inline void cp_zfree(void *ptr, unsigned int len)
{
	kzfree(ptr);
}

/*
 * general functions
 */
int cp_debugfs_init(void);
void cp_debugfs_exit(void);
char *cp_print_status(struct cp_test *test);
int cp_exec_test(struct cp_test *test, unsigned int exectime);
int cp_register_test(struct cp_test *test);

/*
 * functions to register tests
 */
int cp_rng_register(void);
int cp_blkcipher_register(void);
int cp_ablkcipher_register(void);
int cp_shash_register(void);

#endif /* CRYPTOPERF_H */
